# -*- coding: utf-8 -*-
import datetime
import logging
import os
import zipfile
from http import HTTPStatus
from pathlib import Path

import werkzeug
from flask import make_response, send_file, request
from flask_restx import Resource
from sqlalchemy.exc import NoResultFound
from werkzeug.datastructures import FileStorage

from database.models.file_model import FileModel
from database.models.tag_model import TagModel
from search.search import add_to_index, remove_from_index
from webapp import db as DB, basic_auth
from webapp import api
from PIL import Image
import uuid
from datetime import datetime


@api.route("/files/<path:path>", doc={"description": "Download file from server"})
class DownloadFile(Resource):
    @api.response(200, "Success")
    @api.response(404, "NoResultFound Error")
    @api.response(500, "Internal Server Error")
    def get(self, path):
        try:
            if os.path.isfile(f"./files/{path}"):
                return send_file(f"../files/{path}", mimetype='image/png')
            else:
                return make_response({"error": "NoResultFound", "message": "File is not found"}, HTTPStatus.NOT_FOUND)
        except NoResultFound as e:
            return make_response({"error": "NoResultFound", "message": str(e)}, HTTPStatus.NOT_FOUND)
        except Exception as e:
            logging.exception(e)
            return make_response({"error": "Exception", "message": str(e)}, HTTPStatus.INTERNAL_SERVER_ERROR)


parser = api.parser()
parser.add_argument("page", type=int, help="Page Number (Default: 1)")


@api.route("/files/get_all_files", doc={"description": "Get list of all files with pagination"})
@api.expect(parser)
@api.doc(security="basicAuth")
class GetAllFiles(Resource):
    @api.response(200, "Success")
    @api.response(404, "NoResultFound Error")
    @api.response(500, "Internal Server Error")
    @basic_auth.required
    def get(self):
        page = request.args.get("page", 1, type=int)
        try:
            files = DB.session.query(FileModel).paginate(page, per_page=10, error_out=False).items
            return [file.to_dict() for file in files]
        except NoResultFound as e:
            return make_response({"error": "NoResultFound", "message": str(e)}, HTTPStatus.NOT_FOUND)
        except Exception as e:
            logging.exception(e)
            return make_response({"error": "Exception", "message": str(e)}, HTTPStatus.INTERNAL_SERVER_ERROR)


one_file_parser = api.parser()
one_file_parser.add_argument("file_uuid", type=str, help="File UUID")


@api.route("/files/get_one_file", doc={"description": "Get a single file by ID"})
@api.expect(one_file_parser)
@api.doc(security="basicAuth")
class GetOneFile(Resource):
    @api.response(200, "Success")
    @api.response(404, "NoResultFound Error")
    @api.response(500, "Internal Server Error")
    @basic_auth.required
    def get(self):
        file_uuid = request.args.get("file_uuid", '', type=str)
        try:
            file = DB.session.query(FileModel).filter_by(__uuid__=file_uuid).one()
            return [file.to_dict()]
        except NoResultFound as e:
            return make_response({"error": "NoResultFound", "message": str(e)}, HTTPStatus.NOT_FOUND)
        except Exception as e:
            logging.exception(e)
            return make_response({"error": "Exception", "message": str(e)}, HTTPStatus.INTERNAL_SERVER_ERROR)


one_tag_parser = api.parser()
one_tag_parser.add_argument("tag", type=str, help="Tag")


@api.route("/files/get_one_tag", doc={"description": "Get a single TAG"})
@api.expect(one_tag_parser)
@api.doc(security="basicAuth")
class GetOneFile(Resource):
    @api.response(200, "Success")
    @api.response(404, "NoResultFound Error")
    @api.response(500, "Internal Server Error")
    @basic_auth.required
    def get(self):
        tag_name = request.args.get("tag", '', type=str)
        try:
            tags = DB.session.query(TagModel).filter_by(name=tag_name).all()
            return [tag.to_dict() for tag in tags]
        except NoResultFound as e:
            return make_response({"error": "NoResultFound", "message": str(e)}, HTTPStatus.NOT_FOUND)
        except Exception as e:
            logging.exception(e)
            return make_response({"error": "Exception", "message": str(e)}, HTTPStatus.INTERNAL_SERVER_ERROR)


bulk_delete_parser = api.parser()
bulk_delete_parser.add_argument("uuids",
                           required=True,
                           type=str)


@api.route("/files/bulk-delete", doc={"description": "Bulk delete files by UUIDS"})
@api.expect(bulk_delete_parser)
@api.doc(security="basicAuth")
class BulkDeleteFiles(Resource):
    @api.response(200, "Success")
    @api.response(404, "NoResultFound Error")
    @api.response(500, "Internal Server Error")
    @basic_auth.required
    def get(self):
        try:
            uuids = request.args.get("uuids", "", type=str)
            without_white_spaces = uuids.replace(" ", "").replace('"', '').split(',')
            for file_uuid in without_white_spaces:
                delete_file(file_uuid)
            return make_response({"success": "Success", "message": "Files was deleted"}, HTTPStatus.OK)
        except NoResultFound as e:
            return make_response({"error": "NoResultFound", "message": str(e)}, HTTPStatus.NOT_FOUND)
        except Exception as e:
            logging.exception(e)
            return make_response({"error": "Exception", "message": str(e)}, HTTPStatus.INTERNAL_SERVER_ERROR)


@api.route("/files/delete/<path:file_uuid>", doc={"description": "Delete file"})
@api.param("file_uuid", "File UUID")
@api.doc(security="basicAuth")
class DeleteFile(Resource):
    @api.response(200, "Success")
    @api.response(404, "NoResultFound Error")
    @api.response(500, "Internal Server Error")
    @basic_auth.required
    def get(self, file_uuid):
        delete_file(file_uuid)


@api.route("/files/get_all_tags", doc={"description": "Get list of all tags with pagination"})
@api.expect(parser)
@api.doc(security="basicAuth")
class GetAllTags(Resource):
    @api.response(200, "Success")
    @api.response(404, "NoResultFound Error")
    @api.response(500, "Internal Server Error")
    @basic_auth.required
    def get(self):
        page = request.args.get("page", 1, type=int)
        try:
            tags = DB.session.query(TagModel).paginate(page, per_page=100, error_out=False).items
            return [tag.to_dict() for tag in tags]
        except NoResultFound as e:
            return make_response({"error": "NoResultFound", "message": str(e)}, HTTPStatus.NOT_FOUND)
        except Exception as e:
            logging.exception(e)
            return make_response({"error": "Exception", "message": str(e)}, HTTPStatus.INTERNAL_SERVER_ERROR)


upload_parser = api.parser()
upload_parser.add_argument("file",
                           required=True,
                           type=werkzeug.datastructures.FileStorage,
                           location="files")
upload_parser.add_argument("tags",
                           required=False,
                           type=str, action='append')


@api.route("/files/upload", doc={"description": "Upload file to server"})
@api.expect(upload_parser)
@api.doc(security="basicAuth")
class UploadFile(Resource):
    @api.response(200, "Success")
    @api.response(500, "Internal Server Error")
    @basic_auth.required
    def put(self):
        try:
            args = upload_parser.parse_args()
            uploaded_file = args["file"]

            tags = request.args.getlist('tags', type=str)

            now = datetime.now()

            path = f"./files/{now.year}/{now.month}/{now.day}"
            path_for_db = f"files/{now.year}/{now.month}/{now.day}"

            file_uuid = uuid.uuid4()

            filename = f"{file_uuid}.png"
            thumbnail_name = f"thumbnail_{file_uuid}.png"

            image = Image.open(uploaded_file)

            image_thumb = Image.open(uploaded_file)
            image_thumb.thumbnail((90, 90))

            if not Path(path).is_dir():
                try:
                    Path(path).mkdir(parents=True, exist_ok=True)
                except OSError as exc:
                    logging.exception(exc)

            try:
                tag_model_list = set()

                for tag in tags:
                    for splited_tag in tag.split(' '):
                        tag_model = TagModel(name=splited_tag)
                        tag_model_list.add(tag_model)

                new_file = FileModel(filepath=path_for_db, uuid=file_uuid)

                for tag_model in tag_model_list:
                    try:
                        existing = DB.session.query(TagModel).filter_by(name=tag_model.name).one()
                        new_file.__tags__.append(existing)
                    except NoResultFound:
                        DB.session.add(tag_model)
                        new_file.__tags__.append(tag_model)

                DB.session.add(new_file)
                DB.session.commit()
                add_to_index('files', new_file)
                image.save(os.path.join(path, filename))
                image_thumb.save(os.path.join(path, thumbnail_name))
                return {"Status": "Success", "File": filename}, 201
            except Exception as e:
                logging.exception(e)
                return make_response({"error": "Exception", "message": 'IMAGE DO NOT SAVED, TRY AGAIN'}, HTTPStatus.INTERNAL_SERVER_ERROR)

        except Exception as e:
            logging.exception(e)
            return make_response({"error": "Exception", "message": str(e)}, HTTPStatus.INTERNAL_SERVER_ERROR)


@api.route("/files/upload_zip", doc={"description": "Upload zip file to server"})
@api.expect(upload_parser)
@api.doc(security="basicAuth")
class UploadZipFile(Resource):
    @api.response(200, "Success")
    @api.response(500, "Internal Server Error")
    @basic_auth.required
    def put(self):
        try:
            args = upload_parser.parse_args()
            logging.exception(request.headers.get("Authorization"))
            logging.error(request.headers.get("Authorization"))
            file = args["file"]
            file_like_object = file.stream._file
            zipfile_ob = zipfile.ZipFile(file_like_object, 'r')
            file_names = zipfile_ob.namelist()
            file_names = [file_name for file_name in file_names]
            files = [zipfile_ob.open(name) for name in file_names]

            tags = request.args.getlist('tags', type=str)

            now = datetime.now()

            path = f"./files/{now.year}/{now.month}/{now.day}"
            path_for_db = f"files/{now.year}/{now.month}/{now.day}"

            tag_model_list = set()

            for tag in tags:
                for splited_tag in tag.split(' '):
                    tag_model = TagModel(name=splited_tag)
                    try:
                        existing = DB.session.query(TagModel).filter_by(name=tag_model.name).one()
                        tag_model_list.add(existing)
                    except NoResultFound:
                        DB.session.add(tag_model)
                        tag_model_list.add(tag_model)

            for file in files:

                file_uuid = uuid.uuid4()

                filename = f"{file_uuid}.png"
                thumbnail_name = f"thumbnail_{file_uuid}.png"

                image = Image.open(file)

                image_thumb = Image.open(file)
                image_thumb.thumbnail((90, 90))

                if not Path(path).is_dir():
                    try:
                        Path(path).mkdir(parents=True, exist_ok=True)
                    except OSError as exc:
                        print(exc)

                new_file = FileModel(filepath=path_for_db, uuid=file_uuid)

                for tag_model in tag_model_list:
                    new_file.__tags__.append(tag_model)

                DB.session.add(new_file)
                DB.session.commit()
                add_to_index('files', new_file)
                image.save(os.path.join(path, filename))
                image_thumb.save(os.path.join(path, thumbnail_name))

            return {"Status": "Success"}, 201
        except Exception as e:
            logging.exception(e)
            return make_response({"error": "Exception", "message": str(e)}, HTTPStatus.INTERNAL_SERVER_ERROR)


@api.route("/files/search/<path:search_phrase>", doc={"description": "Search"})
@api.expect(parser)
@api.doc(security="basicAuth")
class Search(Resource):
    @api.response(200, "Success")
    @api.response(404, "NoResultFound Error")
    @api.response(500, "Internal Server Error")
    @basic_auth.required
    def get(self, search_phrase):
        page = request.args.get("page", 1, type=int)
        try:
            query, total = FileModel.search(search_phrase, page, 20)
            return [file.to_dict() for file in query]
        except NoResultFound as e:
            return make_response({"error": "NoResultFound", "message": str(e)}, HTTPStatus.NOT_FOUND)
        except Exception as e:
            logging.exception(e)
            return make_response({"error": "Exception", "message": str(e)}, HTTPStatus.INTERNAL_SERVER_ERROR)


def delete_file(file_uuid):
    file = DB.session.query(FileModel).filter_by(__uuid__=file_uuid).one()
    filename = f"{file_uuid}.png"
    thumbnail_name = f"thumbnail_{file_uuid}.png"

    os.remove(f"./{file.filepath}/{filename}")
    os.remove(f"./{file.filepath}/{thumbnail_name}")
    DB.session.delete(file)
    DB.session.commit()
    remove_from_index('files', file)
